from collections import namedtuple
import enum

BERObject = namedtuple('BERObject', ['tag', 'content'])

class IncompleteBERError(ValueError):
	def __init__(self, expected_length=-1):
		super().__init__()
		self.expected_length = expected_length

def decode_ber(data):
	index = 0
	if len(data) < 2:
		raise IncompleteBERError(2)
	identifier = data[index]
	ber_class = identifier >> 6
	ber_constructed = bool(identifier & 0x20)
	ber_type = identifier & 0x1f
	index += 1
	if not data[index] & 0x80:
		length = data[index]
		index += 1
	elif data[index] == 0x80:
		raise ValueError('Indefinite form not implemented')
	elif data[index] == 0xff:
		raise ValueError('BER length invalid')
	else:
		num = data[index] & ~0x80
		index += 1
		if len(data) < index + num:
			raise IncompleteBERError(index + num)
		length = 0
		for octet in data[index:index + num]:
			length = length << 8 | octet
		index += num
	if len(data) < index + length:
		raise IncompleteBERError(index + length)
	ber_content = data[index: index + length]
	rest = data[index + length:]
	return BERObject((ber_class, ber_constructed, ber_type), ber_content), rest

def decode_ber_integer(data):
	if not data:
		return 0
	value = -1 if data[0] & 0x80 else 0
	for octet in data:
		value = value << 8 | octet
	return value

def encode_ber(obj):
	tag = (obj.tag[0] & 0b11) << 6 | (obj.tag[1] & 1) << 5 | (obj.tag[2] & 0b11111)
	length = len(obj.content)
	if length < 127:
		return bytes([tag, length]) + obj.content
	octets = []
	while length:
		octets.append(length & 0xff)
		length = length >> 8
	return bytes([tag, 0x80 | len(octets)]) + bytes(reversed(octets)) + obj.content

def encode_ber_integer(value):
	if value < 0 or value > 255:
		raise NotImplementedError('Encoding of integers greater than 255 is not implemented')
	return bytes([value])

class BERType:
	@classmethod
	def from_ber(cls, data):
		raise NotImplementedError()

	@classmethod
	def to_ber(cls, obj):
		raise NotImplementedError()

	def __bytes__(self):
		return type(self).to_ber(self)

class OctetString(BERType):
	ber_tag = (0, False, 4)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.ber_tag:
			raise ValueError('Expected tag %s but found %s'%(cls.ber_tag, obj.tag))
		return obj.content, rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, bytes):
			raise TypeError()
		return encode_ber(BERObject(cls.ber_tag, obj))

class Integer(BERType):
	ber_tag = (0, False, 2)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.ber_tag:
			raise ValueError()
		return decode_ber_integer(obj.content), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, int):
			raise TypeError()
		return encode_ber(BERObject(cls.ber_tag, encode_ber_integer(obj)))

class Boolean(BERType):
	ber_tag = (0, False, 1)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.ber_tag:
			raise ValueError()
		return bool(decode_ber_integer(obj.content)), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, bool):
			raise TypeError()
		content = b'\xff' if obj else b'\x00'
		return encode_ber(BERObject(cls.ber_tag, content))

class LDAPString(OctetString):
	@classmethod
	def from_ber(cls, data):
		raw, rest = super().from_ber(data)
		return raw.decode(), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, str):
			raise TypeError()
		return super().to_ber(obj.encode())

class LDAPOID(LDAPString):
	pass

class Set(BERType):
	ber_tag = (0, True, 17)
	set_type = OctetString

	@classmethod
	def from_ber(cls, data):
		setobj, rest = decode_ber(data)
		if setobj.tag != cls.ber_tag:
			raise ValueError()
		objs = []
		data = setobj.content
		while data:
			obj, data = cls.set_type.from_ber(data)
			objs.append(obj)
		return list(objs), rest

	@classmethod
	def to_ber(cls, obj):
		content = b''
		for item in obj:
			content += cls.set_type.to_ber(item)
		return encode_ber(BERObject(cls.ber_tag, content))

class SequenceOf(Set):
	ber_tag = (0, True, 16)

class Sequence(BERType):
	ber_tag = (0, True, 16)
	sequence_fields = [
		#(Type, attr_name, default_value, optional?),
	]

	def __init__(self, *args, **kwargs):
		for index, spec in enumerate(type(self).sequence_fields):
			field_type, name, default, optional = spec
			if index < len(args):
				value = args[index]
			elif name in kwargs:
				value = kwargs[name]
			else:
				value = default() if callable(default) else default
			setattr(self, name, value)

	def __repr__(self):
		args = []
		for field_type, name, default, optional in type(self).sequence_fields:
			args.append('%s=%s'%(name, repr(getattr(self, name))))
		return '<%s(%s)>'%(type(self).__name__, ', '.join(args))

	@classmethod
	def from_ber(cls, data):
		seqobj, rest = decode_ber(data)
		if seqobj.tag != cls.ber_tag:
			raise ValueError()
		args = []
		data = seqobj.content
		for field_type, name, default, optional in cls.sequence_fields:
			try:
				obj, data = field_type.from_ber(data)
				args.append(obj)
			except ValueError as e:
				if not optional:
					raise e
				args.append(None)
		return cls(*args), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, cls):
			raise TypeError()
		content = b''
		for field_type, name, default, optional in cls.sequence_fields:
			if not optional or getattr(obj, name) is not None:
				content += field_type.to_ber(getattr(obj, name))
		return encode_ber(BERObject(cls.ber_tag, content))

class Choice(BERType):
	ber_tag = None

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		for subcls in cls.__subclasses__():
			if subcls.ber_tag == obj.tag:
				return subcls.from_ber(data)
		return None, rest

	@classmethod
	def to_ber(cls, obj):
		for subcls in cls.__subclasses__():
			if isinstance(obj, subcls):
				return subcls.to_ber(obj)
		raise TypeError()

class Wrapper(BERType):
	ber_tag = None
	wrapped_attribute = None
	wrapped_type = None
	wrapped_default = None
	wrapped_clsattrs = {}

	def __init__(self, *args, **kwargs):
		cls = type(self)
		attribute = cls.wrapped_attribute
		if args:
			setattr(self, attribute, args[0])
		elif kwargs:
			setattr(self, attribute, kwargs[attribute])
		else:
			setattr(self, attribute, cls.wrapped_default() if callable(cls.wrapped_default) else cls.wrapped_default)

	def __repr__(self):
		return '<%s(%s)>'%(type(self).__name__, repr(getattr(self, type(self).wrapped_attribute)))

	@classmethod
	def from_ber(cls, data):
		class WrappedType(cls.wrapped_type):
			ber_tag = cls.ber_tag
		for key, value in cls.wrapped_clsattrs.items():
			setattr(WrappedType, key, value)
		value, rest = WrappedType.from_ber(data)
		return cls(value), rest

	@classmethod
	def to_ber(cls, obj):
		class WrappedType(cls.wrapped_type):
			ber_tag = cls.ber_tag
		for key, value in cls.wrapped_clsattrs.items():
			setattr(WrappedType, key, value)
		if not isinstance(obj, cls):
			raise TypeError()
		return WrappedType.to_ber(getattr(obj, cls.wrapped_attribute))

def retag(cls, tag):
	class Overwritten(cls):
		ber_tag = tag
	return Overwritten

class AttributeValueAssertion(Sequence):
	sequence_fields = [
		(LDAPString, 'attributeDesc', None, False),
		(OctetString, 'assertionValue', None, False),
	]

class Filter(Choice):
	pass

class FilterAnd(Wrapper, Filter):
	ber_tag = (2, True, 0)
	wrapped_attribute = 'filters'
	wrapped_type = Set
	wrapped_clsattrs = {'set_type': Filter}

class FilterOr(Wrapper, Filter):
	ber_tag = (2, True, 1)
	wrapped_attribute = 'filters'
	wrapped_type = Set
	wrapped_clsattrs = {'set_type': Filter}

class FilterNot(Sequence, Filter):
	ber_tag = (2, True, 2)
	sequence_fields = [
		(Filter, 'filter', None, False)
	]

class FilterEqual(Sequence, Filter):
	ber_tag = (2, True, 3)
	sequence_fields = [
		(LDAPString, 'attribute', None, False),
		(OctetString, 'value', None, False)
	]

class FilterGreaterOrEqual(AttributeValueAssertion, Filter):
	ber_tag = (2, True, 5)

class FilterLessOrEqual(AttributeValueAssertion, Filter):
	ber_tag = (2, True, 6)

class FilterPresent(Wrapper, Filter):
	ber_tag = (2, False, 7)
	wrapped_attribute = 'attribute'
	wrapped_type = LDAPString
	wrapped_default = None

class FilterApproxMatch(AttributeValueAssertion, Filter):
	ber_tag = (2, True, 8)

class FilterExtensibleMatch(Sequence, Filter):
	ber_tag = (2, True, 9)
	sequence_fields = [
		(retag(LDAPString, (2, False, 1)), 'matchingRule', None, True),
		(retag(LDAPString, (2, False, 2)), 'type', None, True),
		(retag(OctetString, (2, False, 3)), 'matchValue', None, False),
		(retag(Boolean, (2, False, 4)), 'dnAttributes', None, True),
	]

class Enum(BERType):
	ber_tag = (0, False, 10)
	enum_type = None

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.ber_tag:
			raise ValueError()
		value = decode_ber_integer(obj.content)
		return cls.enum_type(value), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, cls.enum_type):
			raise TypeError()
		return encode_ber(BERObject(cls.ber_tag, encode_ber_integer(obj.value)))

def wrapenum(enumtype):
	class WrappedEnum(Enum):
		enum_type = enumtype
	return WrappedEnum

class SearchScope(enum.Enum):
	baseObject = 0 # The scope is constrained to the entry named by baseObject.
	singleLevel = 1 # The scope is constrained to the immediate subordinates of the entry named by baseObject.
	wholeSubtree = 2 # The scope is constrained to the entry named by baseObject and to all its subordinates.

class DerefAliases(enum.Enum):
	neverDerefAliases = 0
	derefInSearching = 1
	derefFindingBaseObj = 2
	derefAlways = 3

class LDAPResultCode(enum.Enum):
	success                      = 0
	operationsError              = 1
	protocolError                = 2
	timeLimitExceeded            = 3
	sizeLimitExceeded            = 4
	compareFalse                 = 5
	compareTrue                  = 6
	authMethodNotSupported       = 7
	strongerAuthRequired         = 8
	# -- 9 reserved --
	referral                     = 10
	adminLimitExceeded           = 11
	unavailableCriticalExtension = 12
	confidentialityRequired      = 13
	saslBindInProgress           = 14
	noSuchAttribute              = 16
	undefinedAttributeType       = 17
	inappropriateMatching        = 18
	constraintViolation          = 19
	attributeOrValueExists       = 20
	invalidAttributeSyntax       = 21
	# -- 22-31 unused --
	noSuchObject                 = 32
	aliasProblem                 = 33
	invalidDNSyntax              = 34
	# -- 35 reserved for undefined isLeaf --
	aliasDereferencingProblem    = 36
	# -- 37-47 unused --
	inappropriateAuthentication  = 48
	invalidCredentials           = 49
	insufficientAccessRights     = 50
	busy                         = 51
	unavailable                  = 52
	unwillingToPerform           = 53
	loopDetect                   = 54
	# -- 55-63 unused --
	namingViolation              = 64
	objectClassViolation         = 65
	notAllowedOnNonLeaf          = 66
	notAllowedOnRDN              = 67
	entryAlreadyExists           = 68
	objectClassModsProhibited    = 69
	# -- 70 reserved for CLDAP --
	affectsMultipleDSAs          = 71
	# -- 72-79 unused --
	other                        = 80

class LDAPResult(Sequence):
	ber_tag = (5, True, 1)
	sequence_fields = [
		(wrapenum(LDAPResultCode), 'resultCode', None, False),
		(LDAPString, 'matchedDN', '', False),
		(LDAPString, 'diagnosticMessage', '', False),
	]

class AttributeSelection(SequenceOf):
	set_type = LDAPString

class AuthenticationChoice(Choice):
	pass

class SimpleAuthentication(Wrapper, AuthenticationChoice):
	ber_tag = (2, False, 0)
	wrapped_attribute = 'password'
	wrapped_type = OctetString
	wrapped_default = b''

	def __repr__(self):
		if not self.password:
			return '<%s(EMPTY PASSWORD)>'%(type(self).__name__)
		return '<%s(PASSWORD HIDDEN)>'%(type(self).__name__)

class SaslCredentials(Sequence, AuthenticationChoice):
	ber_tag = (2, True, 3)
	sequence_fields = [
		(LDAPString, 'mechanism', None, False),
		(OctetString, 'credentials', None, True),
	]

class AttributeValueSet(Set):
	set_type = OctetString

class PartialAttribute(Sequence):
	sequence_fields = [
		(LDAPString, 'type', None, False),
		(AttributeValueSet, 'vals', lambda: [], False),
	]

class PartialAttributeList(SequenceOf):
	set_type = PartialAttribute

class Attribute(Sequence):
	# Constrain: vals must not be empty
	sequence_fields = [
		(LDAPString, 'type', None, False),
		(AttributeValueSet, 'vals', lambda: [], False),
	]

class AttributeList(SequenceOf):
	set_type = Attribute

class ProtocolOp(Choice):
	pass

class BindRequest(Sequence, ProtocolOp):
	ber_tag = (1, True, 0)
	sequence_fields = [
		(Integer, 'version', 3, False),
		(LDAPString, 'name', '', False),
		(AuthenticationChoice, 'authentication', lambda: SimpleAuthentication(), False)
	]

class BindResponse(Sequence, ProtocolOp):
	ber_tag = (1, True, 1)
	sequence_fields = [
		(wrapenum(LDAPResultCode), 'resultCode', None, False),
		(LDAPString, 'matchedDN', '', False),
		(LDAPString, 'diagnosticMessage', '', False),
		(retag(OctetString, (2, False, 7)), 'serverSaslCreds', None, True)
	]

class UnbindRequest(Sequence, ProtocolOp):
	ber_tag = (1, False, 2)

class SearchRequest(Sequence, ProtocolOp):
	ber_tag = (1, True, 3)
	sequence_fields = [
		(LDAPString, 'baseObject', '', False),
		(wrapenum(SearchScope), 'scope', SearchScope.wholeSubtree, False),
		(wrapenum(DerefAliases), 'derefAliases', DerefAliases.neverDerefAliases, False),
		(Integer, 'sizeLimit', 0, False),
		(Integer, 'timeLimit', 0, False),
		(Boolean, 'typesOnly', False, False),
		(Filter, 'filter', lambda: FilterPresent('objectClass'), False),
		(AttributeSelection, 'attributes', lambda: [], False)
	]

	@classmethod
	def from_ber(cls, data):
		return super().from_ber(data)

class SearchResultEntry(Sequence, ProtocolOp):
	ber_tag = (1, True, 4)
	sequence_fields = [
		(LDAPString, 'objectName', '', False),
		(PartialAttributeList, 'attributes', lambda: [], False),
	]

class SearchResultDone(LDAPResult, ProtocolOp):
	ber_tag = (1, True, 5)

class ModifyOperation(enum.Enum):
	add = 0
	delete = 1
	replace = 2

class ModifyChange(Sequence):
	sequence_fields = [
		(wrapenum(ModifyOperation), 'operation', None, False),
		(PartialAttribute, 'modification', None, False),
	]

class ModifyChanges(SequenceOf):
	set_type = ModifyChange

class ModifyRequest(Sequence, ProtocolOp):
	ber_tag = (1, True, 6)
	sequence_fields = [
		(LDAPString, 'object', None, False),
		(ModifyChanges, 'changes', None, False),
	]

class ModifyResponse(LDAPResult, ProtocolOp):
	ber_tag = (1, True, 7)

class AddRequest(Sequence, ProtocolOp):
	ber_tag = (1, True, 8)
	sequence_fields = [
		(LDAPString, 'entry', None, False),
		(AttributeList, 'attributes', None, False),
	]

class AddResponse(LDAPResult, ProtocolOp):
	ber_tag = (1, True, 9)

class DelRequest(Wrapper, ProtocolOp):
	ber_tag = (1, False, 10)
	wrapped_attribute = 'dn'
	wrapped_type = LDAPString
	wrapped_default = None

class DelResponse(LDAPResult, ProtocolOp):
	ber_tag = (1, True, 11)

class ModifyDNRequest(Sequence, ProtocolOp):
	ber_tag = (1, True, 12)
	sequence_fields = [
		(LDAPString, 'entry', None, False),
		(LDAPString, 'newrdn', None, False),
		(Boolean, 'deleteoldrdn', None, False),
		(retag(LDAPString, (2, False, 0)), 'newSuperior', None, True),
	]

class ModifyDNResponse(LDAPResult, ProtocolOp):
	ber_tag = (1, True, 13)

class CompareRequest(Sequence, ProtocolOp):
	ber_tag = (1, True, 14)
	sequence_fields = [
		(LDAPString, 'entry', None, False),
		(AttributeValueAssertion, 'ava', None, False),
	]

class CompareResponse(LDAPResult, ProtocolOp):
	ber_tag = (1, True, 15)

class AbandonRequest(Wrapper, ProtocolOp):
	ber_tag = (1, False, 16)
	wrapped_attribute = 'messageID'
	wrapped_type = Integer
	wrapped_default = None

class ExtendedRequest(Sequence, ProtocolOp):
	ber_tag = (1, True, 23)
	sequence_fields = [
		(retag(LDAPOID, (2, False, 0)), 'requestName', None, True),
		(retag(OctetString, (2, False, 1)), 'requestValue', None, True),
	]

class ExtendedResponse(Sequence, ProtocolOp):
	ber_tag = (1, True, 24)
	sequence_fields = [
		(wrapenum(LDAPResultCode), 'resultCode', None, False),
		(LDAPString, 'matchedDN', '', False),
		(LDAPString, 'diagnosticMessage', '', False),
		(retag(LDAPOID, (2, False, 10)), 'responseName', None, True),
		(retag(OctetString, (2, False, 11)), 'responseValue', None, True),
	]

class IntermediateResponse(Sequence, ProtocolOp):
	ber_tag = (1, True, 25)
	sequence_fields = [
		(retag(LDAPOID, (2, False, 0)), 'responseName', None, True),
		(retag(OctetString, (2, False, 1)), 'responseValue', None, True),
	]

class Control(Sequence):
	sequence_fields = [
		(LDAPOID, 'controlType', None, False),
		(Boolean, 'criticality', None, True),
		(OctetString, 'controlValue', None, True),
	]

class Controls(SequenceOf):
	ber_tag = (2, True, 0)
	set_type = Control

class LDAPMessage(Sequence):
	sequence_fields = [
		(Integer, 'messageID', None, False),
		(ProtocolOp, 'protocolOp', None, False),
		(Controls, 'controls', None, True)
	]

class ShallowLDAPMessage(BERType):
	ber_tag = (0, True, 16)

	def __init__(self, messageID=None, protocolOpType=None, data=None):
		self.messageID = messageID
		self.protocolOpType = protocolOpType
		self.data = data

	def decode(self):
		return LDAPMessage.from_ber(self.data)

	@classmethod
	def from_ber(cls, data):
		seq, rest = decode_ber(data)
		data = data[:len(data)-len(rest)]
		if seq.tag != cls.ber_tag:
			raise ValueError()
		content = seq.content
		messageID, content = Integer.from_ber(content)
		op, content = decode_ber(content)
		for subcls in ProtocolOp.__subclasses__():
			if subcls.ber_tag == op.tag:
				return cls(messageID, subcls, data), rest
		return cls(messageID, None, data), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, cls):
			raise TypeError()
		return obj.data

# Extended Operation Values

EXT_STARTTLS_OID = '1.3.6.1.4.1.1466.20037'
EXT_WHOAMI_OID = '1.3.6.1.4.1.4203.1.11.3'
EXT_PASSWORD_MODIFY_OID = '1.3.6.1.4.1.4203.1.11.1'

class PasswdModifyRequestValue(Sequence):
	sequence_fields = [
		(retag(LDAPString, (2, False, 0)), 'userIdentity', None, True),
		(retag(OctetString, (2, False, 1)), 'oldPasswd', None, True),
		(retag(OctetString, (2, False, 2)), 'newPasswd', None, True),
	]

class PasswdModifyResponseValue(Sequence):
	sequence_fields = [
		(retag(OctetString, (2, False, 0)), 'genPasswd', None, True),
	]

bind1 = b'0\x0c\x02\x01\x01`\x07\x02\x01\x03\x04\x00\x80\x00'
bind2 = b'0\x1c\x02\x01\x01`\x17\x02\x01\x03\x04\nuid=foobar\x80\x06foobar'
search1 = b'0?\x02\x01\x02c:\x04\x1aou=users,dc=example,dc=com\n\x01\x01\n\x01\x00\x02\x01\x00\x02\x01\x00\x01\x01\x00\x87\x0bobjectclass0\x00'
search2 = b'0@\x02\x01\x02c;\x04\x00\n\x01\x02\n\x01\x00\x02\x01\x00\x02\x01\x00\x01\x01\x00\xa0&\xa3\x12\x04\x0bobjectClass\x04\x03top\xa3\x10\x04\x03uid\x04\ttestadmin0\x00'
unbind1 = b'0\x05\x02\x01\x03B\x00'

